#include "config.h"
//-----------------

#include "Packet.h"

//-----------------------------------------------------------------------------

inline unsigned char COMP_ID(const unsigned char id) { return id & 0x0F; };

const unsigned char MSG_CTS        = 0x09;  // clear to send
const unsigned char MSG_ACK        = 0x56;
const unsigned char MSG_NACK       = 0x5A;  // retransmit last packet
const unsigned char MSG_PING       = 0x68;
const unsigned char MSG_PRTSCR     = 0x6D;  // print screen
const unsigned char MSG_MORE       = 0x78;  // more vars follow
const unsigned char MSG_KEY        = 0x87;  // send keystroke
const unsigned char MSG_EOT        = 0x92;  // end of transmission

const unsigned char PKT_HEADER     = 0x06;
const unsigned char PKT_DATA       = 0x15;
const unsigned char PKT_ABORT      = 0x36;  // "Skip" or "Exit"
  const unsigned char DATA_EXIT    = 0x01;
  const unsigned char DATA_SKIP    = 0x02;
  const unsigned char DATA_MEM     = 0x05;  // no mem to store var
const unsigned char PKT_REQ_HEADER = 0xA2;  // request variable
const unsigned char PKT_SEND_HEADER= 0xC9;  // header for "send(x)" (TI-82)

const unsigned typetype_map[256] = {
  // bit 0: 1=valid message or packet
  // bit 1: 1=data field (and checksum) required (PKT, not MSG)
//0 1 2 3 4 5 6 7  8 9 A B C D E F 10 1 2 3 4 5 6 7 18 9 A B C D E F
  0,0,0,0,0,0,3,0, 0,1,0,0,0,0,0,0, 0,0,0,0,0,3,0,0, 0,0,0,0,0,0,0,0, /*00-1F*/
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,3,0, 0,0,0,0,0,0,0,0, /*20-3F*/
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,1,0, 0,0,1,0,0,0,0,0, /*40-5F*/
  0,0,0,0,0,0,0,0, 1,0,0,0,0,1,0,0, 0,0,0,0,0,0,0,0, 1,0,0,0,0,0,0,0, /*60-7F*/
  0,0,0,0,0,0,0,1, 0,0,0,0,0,0,0,0, 0,0,1,0,0,0,0,0, 0,0,0,0,0,0,0,0, /*80-9F*/
  0,0,3,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, /*A0-BF*/
  0,0,0,0,0,0,0,0, 0,3,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, /*C0-DF*/
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0  /*E0-FF*/
};

inline bool hasData(const unsigned char n) { return typetype_map[n] & 2; };
inline bool isMSG(const unsigned char n) { return (typetype_map[n]==1); };
inline bool isPKT(const unsigned char n) { return (typetype_map[n]==3); };

//-----------------------------------------------------------------------------

/*
class Msg {
  char buf[4];
public:
  Msg(unsigned char type, unsigned int msg) {
    buf[0] = 0;
    buf[1] = type;
    buf[3] = msg & 0xFF;
    buf[2] = (msg >> 8) & 0xFF;
  }
  friend bool Send(unsigned char id, Msg &M);
  friend bool Get(Msg &M);
};
*/

//----- Debugging function ---------------------------------------------------

#ifdef SYSLOG_DEBUG
void log_packet(unsigned char id, unsigned char type, unsigned length)
{
    char *s, *dir, typebuf[3];

    if (id & 0xF0)
      dir = "<-";
    else
      dir = "->";

    switch (type) {
    case PKT_HEADER     : s = "HEADER"; break;
    case MSG_CTS        : s = "CTS"; break;
    case PKT_DATA       : s = "DATA"; break;
    case PKT_ABORT      : s = "ABORT"; break;
    case MSG_ACK        : s = "ACK"; break;
    case MSG_NACK       : s = "NACK"; break;
    case MSG_PING       : s = "PING"; break;
    case MSG_PRTSCR     : s = "PRTSCR"; break;
    case MSG_MORE       : s = "MORE"; break;
    case MSG_KEY        : s = "KEY"; break;
    case MSG_EOT        : s = "EOT"; break;
    case PKT_REQ_HEADER : s = "REQ_HEADER"; break;
    case PKT_SEND_HEADER: s = "SEND_HEADER"; break;
    default:
        sprintf(typebuf, "%02X", type);
        s = typebuf;
    }
    if (hasData(type))
        syslog(LOG_DEBUG, "%02X%s%-12s Len.=%04X", id, dir, s, length);
    else
        syslog(LOG_DEBUG, "%02X%s%-12s Data=%04X", id, dir, s, length);
}
#endif // SYSLOG_DEBUG

//-----------------------------------------------------------------------------

Packet::PacketError get(Packet& P, FILE *dev) {
  return P.get(dev);
}

Packet::PacketError send(const Packet& P, FILE *dev) {
  return P.send(dev);
}

void Packet::Init(unsigned char nid, unsigned char ntype)
{
  if (delete_data)
    delete [] data;
  delete_data = false;

  id = nid;
  type = ntype;
  length = 0;
  data = 0;
}

void Packet::Init(unsigned char nid, unsigned char ntype,
		  int nlength, unsigned char *ndata)
{
  if (delete_data)
    delete [] data;
  data = ndata;
  delete_data = false;

  id = nid;
  type = ntype;
  length = nlength;
}

// return true if there is a packet, false on error
// does not send a response
Packet::PacketError Packet::_get(FILE *dev)
{
  int c;

  // blow away old buffer
  if (delete_data) {
    delete [] data;
    data = 0;
    delete_data = false;
  }

  do {
    // read calculator ID
    if ((c = fgetc(dev)) == EOF)
      return E_TIMEOUT;
    id = (unsigned char)c;
  } while (id!=ID82 && id!=ID83 && id!=ID85 &&
	   id!=ID86 && id!=ID89 && id!=ID92);

  // read packet type
  if ((c = fgetc(dev)) == EOF)
    return E_TIMEOUT;
  type = (unsigned char)c;

  // read data length
  if ((c = fgetc(dev)) == EOF)
    return E_TIMEOUT;
  length = (unsigned)c;
  if ((c = fgetc(dev)) == EOF)
    return E_TIMEOUT;
  length += (unsigned)(c<<8);

#ifdef SYSLOG_DEBUG
  log_packet(id, type, length);
#endif

  // if the packet is a message, no data will follow, so exit
  if (!hasData(type))
    return E_OK;

  // allocate data space
  data = new unsigned char[length];

  // read packet data
  if (length != fread(data, sizeof(char), length, dev)) {
    delete [] data;
    length = 0;
    return E_TIMEOUT;
  }

  // read checksum
  unsigned remote_checksum;
  if ((c = fgetc(dev)) == EOF) {
    delete [] data;
    length = 0;
    return E_TIMEOUT;
  }
  remote_checksum = (unsigned)c;
  if ((c = fgetc(dev)) == EOF) {
    delete [] data;
    length = 0;
    return E_TIMEOUT;
  }
  remote_checksum += (unsigned)(c<<8);

  // At this point, the packet has been received, although
  // it may be in error.  Keep the data in either case.
  delete_data = true;

  // calculate checksum
  unsigned local_checksum = 0;
  for (unsigned i = 0; i < length; i++)
    local_checksum += data[i];
  local_checksum &= 0xFFFF;

  // compare checksums
  if (local_checksum != remote_checksum)
  {
#ifdef SYSLOG_DEBUG
      syslog(LOG_DEBUG, "    Checksum error: local=%04X remote=%04X",
	     local_checksum, remote_checksum);
#endif
      return E_CHECKSUM;
  }

  return E_OK;
}

// if there is a packet, send ACK; on checksum error, send NACK
Packet::PacketError Packet::get(FILE *dev)
{
  Packet packet(0, MSG_ACK);
  PacketError err = _get(dev);

  switch(err) {
  case E_OK:
    packet.Init(COMP_ID(id), MSG_ACK);
    packet._send(dev);
    return E_OK;
  case E_TIMEOUT:
    return E_TIMEOUT;
  case E_CHECKSUM:
    packet.Init(COMP_ID(id), MSG_NACK);
    packet._send(dev);
    return E_CHECKSUM;
  default:
    return E_PROTOCOL;
  }
}

// does not wait for ACK
Packet::PacketError Packet::_send(FILE *dev) const
{
#ifdef SYSLOG_DEBUG
  log_packet(id, type, length);
#endif

  if (fputc(id, dev) != id)
    return E_TIMEOUT;

  if (fputc(type, dev) != type)
    return E_TIMEOUT;

  // write length
  if (fputc((length&0xFF), dev) != (length&0xFF))
    return E_TIMEOUT;

  if (fputc((length>>8), dev) != (length>>8))
    return E_TIMEOUT;

  // if the packet is a message, no data will follow, so exit
  // also exit if *data is null
  if (!hasData(type) || (data==0))
    return E_OK;

  // write data
  if (length != fwrite(data, sizeof(char), length, dev))
    return E_TIMEOUT;
  
  // calculate checksum
  unsigned local_checksum = 0;
  for (unsigned i = 0; i < length; i++)
    local_checksum += data[i];
  local_checksum &= 0xFFFF;

#ifdef SYSLOG_DEBUG
      syslog(LOG_DEBUG, "    Checksum: local=%04X", local_checksum);
#endif

  int c = local_checksum & 0xFF;
  if (fputc(c, dev) != c)
    return E_TIMEOUT;
  c = (local_checksum>>8) & 0xFF;
  if (fputc(c, dev) != c)
    return E_TIMEOUT;

  return E_OK;
}

Packet::PacketError Packet::send(FILE *dev) const
{
  Packet packet;
  PacketError err;

  do {
    err = _send(dev);        // send data
    if (err != E_OK)
      return err;

    err = packet._get(dev);  // wait for ACK
    if (err != E_OK) {
      if (packet.type == PKT_ABORT) // some skip packets are strange
	return E_REJ;
      return err;
    }

    if (packet.type == MSG_ACK)
      return E_OK;
    else if (packet.type == PKT_ABORT)
      return E_REJ;
    else if (packet.type != MSG_NACK)
      return E_PROTOCOL;

  } while (packet.type == MSG_NACK);

  return E_PROTOCOL;
}
