#ifndef __PACKET_H__
#define __PACKET_H__

#include <stdio.h>
#include "calcids.h"

//-----------------------------------------------------------------------------

unsigned char COMP_ID(const unsigned char id);

extern const unsigned char MSG_CTS;         // clear to send
extern const unsigned char MSG_ACK;
extern const unsigned char MSG_NACK;        // retransmit last packet
extern const unsigned char MSG_PING;
extern const unsigned char MSG_PRTSCR;      // print screen
extern const unsigned char MSG_MORE;        // more vars follow
extern const unsigned char MSG_KEY;         // send keystroke
extern const unsigned char MSG_EOT;         // end of transmission

extern const unsigned char PKT_HEADER;
extern const unsigned char PKT_DATA;
extern const unsigned char PKT_ABORT;       // "Skip" or "Exit"
  extern const unsigned char DATA_EXIT;
  extern const unsigned char DATA_SKIP;
  extern const unsigned char DATA_MEM;       // skip for lack of memory
extern const unsigned char PKT_REQ_HEADER;  // request variable
extern const unsigned char PKT_SEND_HEADER; // header for "send(x)" (TI-82)

//-----------------------------------------------------------------------------

class Packet {
  bool delete_data;
  unsigned char *data;
  // if "data" is 0, it means the packet type is a MSG_* (message)
  // therefore, no data or checksum will be sent with send().
public:
  enum PacketError { E_OK=0, E_REJ, E_TIMEOUT, E_CHECKSUM, E_PROTOCOL };
  unsigned char id;
  unsigned char type;
  int length;

  void Init(unsigned char nid=0, unsigned char ntype=0);
  void Init(unsigned char nid, unsigned char ntype,
	    int nlength, unsigned char *data=0);

  Packet(unsigned char nid=0, unsigned char ntype=0)
    {
      delete_data = false;
      Init(nid,ntype);
    }
  Packet(unsigned char nid, unsigned char ntype,
	 int nlength, void *ndata=0)
    {
      delete_data = false;
      Init(nid,ntype,nlength,(unsigned char*)ndata);
    }
  Packet(Packet& P)
    : id(P.id), type(P.type), length(P.length), delete_data(true)
    {
      data = new unsigned char[length];
      memcpy((void*)data, (void*)P.data, length);
    }

  ~Packet() { if (delete_data) delete [] data; }

  unsigned char& operator[] (int index) { return data[index]; }
  PacketError _get(FILE *dev);
  PacketError get(FILE *dev);
  friend PacketError get(Packet& P, FILE *dev);
  PacketError _send(FILE *dev) const;
  PacketError send(FILE *dev) const;
  friend PacketError send(const Packet& P, FILE *dev);
};

#endif // __PACKET_H__
